/* SPDX-License-Identifier: GPL-2.0-only */

#ifndef __ASM_INSN_DEF_H
#define __ASM_INSN_DEF_H

#include <linux/stringify.h>
#include <asm/gpr-num.h>
#include <asm/asm.h>

#define INSN_STR(x)		__stringify(x)
#define CSR_RD_SHIFT		0
#define CSR_RJ_SHIFT		5
#define CSR_SIMM14_SHIFT	10
#define CSR_OPCODE_SHIFT	24

#define DEFINE_INSN_CSR							\
	__DEFINE_ASM_GPR_NUMS						\
"	.macro insn_csr, opcode, rj, rd, simm14\n"			\
"	.4byte	((\\opcode << " INSN_STR(CSR_OPCODE_SHIFT) ") |"	\
"		 (.L__gpr_num_\\rj << " INSN_STR(CSR_RJ_SHIFT) ") |"	\
"		 (.L__gpr_num_\\rd << " INSN_STR(CSR_RD_SHIFT) ") |"	\
"		 (\\simm14 << " INSN_STR(CSR_SIMM14_SHIFT) "))\n"	\
"	.endm\n"

#define UNDEFINE_INSN_CSR						\
"	.purgem insn_csr\n"

#define __INSN_CSR(opcode, rj, rd, simm14)				\
	DEFINE_INSN_CSR							\
	"insn_csr " opcode ", " rj ", " rd ", " simm14 "\n"		\
	UNDEFINE_INSN_CSR


#define INSN_CSR(opcode, rj, rd, simm14)				\
	__INSN_CSR(LARCH_##opcode, LARCH_##rj, LARCH_##rd,		\
		   LARCH_##simm14)

#define __ASM_STR(x)		#x
#define LARCH_OPCODE(v)		__ASM_STR(v)
#define LARCH_SIMM14(v)		__ASM_STR(v)
#define __LARCH_REG(v)		__ASM_STR(v)
#define LARCH___RD(v)		__LARCH_REG(v)
#define LARCH___RJ(v)		__LARCH_REG(v)
#define LARCH_OPCODE_GCSR	LARCH_OPCODE(5)

#define GCSR_read(csr, rd)						\
	INSN_CSR(OPCODE_GCSR, __RJ(zero), __RD(rd), SIMM14(csr))

#define GCSR_write(csr, rd)						\
	INSN_CSR(OPCODE_GCSR, __RJ($r1), __RD(rd), SIMM14(csr))

#define GCSR_xchg(csr, rj, rd)						\
	INSN_CSR(OPCODE_GCSR, __RJ(rj), __RD(rd), SIMM14(csr))

#endif /* __ASM_INSN_DEF_H */
